function [] = Preprocess_4_FillHoles(base_dir)
%FILLHOLES Fill holes in annotation

  input_dir = [base_dir 'processed3/'];
  output_dir = [base_dir 'processed4/'];

  THRESHOLD = 0.1;
  CLOSE_AMOUNT = 5;
  
  % list all .png files. A filled version will be created for each of these
  filenames = dir([input_dir '*.png']);
  filenames = {filenames.name};
  
  for f = filenames
    img = imread([input_dir f{1}]);
    % Check if pixel is red (red channel is almost 1 and rest are almost 0)
    r_on = img(:, :, 1) > 1 - THRESHOLD;
    g_off = img(:, :, 2) < THRESHOLD;
    b_off = img(:, :, 3) < THRESHOLD;
    red_pixels = r_on & g_off & b_off;
    
    % Close a bit and fill holes, including boundary holes
    se = strel('disk', CLOSE_AMOUNT);
    fixed_pixels = imclose(red_pixels, se);
    % Add a line of pixels on the boundary of the image, where appropriate. 
    % This enables imfill to fill those areas.
    fixed_pixels = CloseBoundaryLoops(fixed_pixels);
    % fill holes
    fixed_pixels = imfill(fixed_pixels, 'holes');
    % Convert to 3 channels (as the red channel)
    fixed_pixels = double(padarray(fixed_pixels, [0 0 2], 'post'));
    
    % Save filled image
    imwrite(fixed_pixels, [output_dir f{1}], 'Transparency', [0 0 0]);
  end

end

function [bw_out] = CloseBoundaryLoops(bw_in)
  % We close a loop if a connected component has two contact points with
  % the same edge of the image or with adjacent edges of the image (but not
  % opposing)
  bw_out = bw_in;
  stats = regionprops(bw_in, 'Extrema');
  [bw_height, bw_width] = size(bw_in);
  
  for ii = 1:numel(stats)
    % top-top
    if (stats(ii).Extrema(1:2, 2) < 1)
      [bw_out] = FillSide(bw_out, 'top', ceil(stats(ii).Extrema(1, 1)), floor(stats(ii).Extrema(2, 1)));
    end
    
    % left-left
    if (stats(ii).Extrema(7:8, 1) < 1)
      [bw_out] = FillSide(bw_out, 'left', ceil(stats(ii).Extrema(8, 2)), floor(stats(ii).Extrema(7, 2)));
    end
    
    % right-right
    if (stats(ii).Extrema(3:4, 1) > bw_width)
      [bw_out] = FillSide(bw_out, 'right', ceil(stats(ii).Extrema(3, 2)), floor(stats(ii).Extrema(4, 2)));
    end
    
    % bottom-bottom
    if (stats(ii).Extrema(5:6, 2) > bw_height)
      [bw_out] = FillSide(bw_out, 'bottom', ceil(stats(ii).Extrema(6, 1)), floor(stats(ii).Extrema(5, 1)));
    end
  end
end

function [bw_out] = FillSide(bw_in, side, from_ind, to_ind)
  bw_out = bw_in;
  [bw_height, bw_width] = size(bw_in);
  
  switch side
    case 'top'
      bw_out(1, from_ind:to_ind) = true;
    case 'left'
      bw_out(from_ind:to_ind, 1) = true;
    case 'bottom'
      bw_out(bw_height, from_ind:to_ind) = true;
    case 'right'
      bw_out(from_ind:to_ind, bw_width) = true;
    otherwise
      error('Invalid option');
  end
end
